<?php
namespace PrestaShop\Module\LCVPrestaConnector;

use Exception;
use StockAvailable;

/**
 * Synchronisation des stocks
 */
class SyncerStocks
{
    /**
     * Synchroniseur
     * 
     */
    private Syncer $syncer;

    /**
     * Constructeur
     * 
     * @param Syncer $syncer Synchroniseur
     */
    public function __construct(Syncer $syncer)
    {
        $this->syncer = $syncer;
    }

    /**
     * Synchronisation des stocks
     */
    public function sync()
    {
        $bridge = $this->syncer->bridge;

        $startTime = microtime(true);
        $nbRefsStocks = 0;
        
        // On récupère les stocks depuis le pont, en boucle
        while ($stocks = $bridge->pullStocks())            
            $nbRefsStocks += $this->syncStocks($stocks);

        $stopTime = microtime(true);
        if ($nbRefsStocks > 0)
            $this->syncer->audit(sprintf("Synchronisation des stocks de %d références effectuée en %.2f secondes", $nbRefsStocks, $stopTime - $startTime));    
    }

    /**
     * Synchronisation des stocks précisés
     * 
     * @param array<Stock> $stocks Stock à synchroniser
     * @return int Nombre de stocks synchronisés 
     */
    public function syncStocks(array & $stocks) : int
    {
        $cfg = $this->syncer->module->getCfg();

        // Identifions les références concernées
        $refMngr = $this->syncer->getReferencesManager();

        $noStockVisibility = $cfg->get(SyncConfiguration::CFG_NO_STOCK_PRODUCT_VISIBILITY);
        if (empty($noStockVisibility))
            $noStockVisibility = null;

        $securityStock = $cfg->get(SyncConfiguration::CFG_SECURED_STOCK);
        if (!$securityStock)
            $securityStock = 0;

        // Un tableau de stock par id_product_attribute
        $refs = [];
        $nbDone = 0;

        foreach ($stocks as $stock)
        {
            foreach ($stock->refs as $ref)
            {
                $combination = $refMngr->translateCombinationRefs($ref);
                if ($combination)
                {
                    foreach ($combination as $c)
                    {
                        // Trouvé !                
                        $stock->id_product = $c[0]; // Pour simplifier la suite
                        $refs[$c[1]] = $stock;
                    }
                }
            }                    
        }

        if (count($refs) == 0)
            return 0;   // Pas de stock à synchroniser

        // Chargeons les stocks concernés depuis la table des stocks
        $currentStocks = [];
        $db = \Db::getInstance();
        $sql = 'SELECT id_product_attribute, store_code, qty FROM `'.$this->syncer->module->TblPrefix.'stocks` WHERE `id_product_attribute` IN ('.implode(',', array_map('intval', array_keys($refs))).')';
        $res = $db->executeS($sql);
        if ($res === FALSE)
            throw new Exception('Erreur lors de la récupération des stocks par magasin : '.$db->getMsgError());
        foreach ($res as $row)
        {
            $id_product_attribute = (int) $row['id_product_attribute'];
            $qty = (int) $row['qty'];

            $currentStocks[$id_product_attribute][$row['store_code']] = $qty;
        }

        $ctx = \Context::getContext();
        // Pays du magasin
        $id_country = $ctx->country->id;
        
        // Chargeons les stocks des produits actuels depuis la table des stocksAvailable
        // dans une structure $id_product => [ visibility => 'both|search...', stock => [id_product_attribute => qty] ]        
        $currentSituation = [];
        $sql = 'SELECT a.`id_product`, a.`id_product_attribute`, a.`quantity`, ps.`visibility`, ps.`price`, pas.`price` AS pa_price, ';
        // Dans la requête, on calcul le taux de tva depuis ps.id_tax_rules_group
        $sql .= '(SELECT t.rate FROM `'._DB_PREFIX_.'tax` AS t LEFT JOIN `'._DB_PREFIX_.'tax_rule` as r ON (t.`id_tax` = r.`id_tax`) WHERE r.`id_tax_rules_group` = ps.`id_tax_rules_group` AND r.`id_country` = '.$id_country.' LIMIT 1) AS tax_rate';
        $sql .= ' FROM `'._DB_PREFIX_.'stock_available` a';
        // On jointe avec la table product pour avoir la visibilité
        $sql .= ' INNER JOIN `'._DB_PREFIX_.'product` p ON (p.`id_product` = a.`id_product`)';
        // Jointure avec la table product_shop
        $sql .= ' INNER JOIN `'._DB_PREFIX_.'product_shop` ps ON (ps.`id_product` = a.`id_product` AND ps.`id_shop` = '.(int) $ctx->shop->id.')';
        // Jointure avec la table product_attribute_shop
        $sql .= ' INNER JOIN `'._DB_PREFIX_.'product_attribute_shop` pas ON (pas.`id_product_attribute` = a.`id_product_attribute` AND pas.`id_shop` = '.(int) $ctx->shop->id.')';
        $sql .= ' WHERE a.`id_product_attribute` IN ('.implode(',', array_map('intval', array_keys($refs))).')';
        $res = $db->executeS($sql);
        if ($res === FALSE)
            throw new Exception('Erreur lors de la récupération des stocks StockAvailable : '.$db->getMsgError());
        foreach ($res as $row)
        {
            $id_product = (int) $row['id_product'];
            $id_product_attribute = (int) $row['id_product_attribute'];
            $qty = (int) $row['quantity'];
            $visibility = $row['visibility'];
            $productPrice = $row['price'];
            $attrProductPrice = $row['pa_price'];

            if (!isset($currentSituation[$id_product]))
                $currentSituation[$id_product] = [
                    'visibility' => $visibility, 
                    'price' => $productPrice,
                    'tax_rate' => $row['tax_rate'] ?? 1,
                    'attr_prices' => [],
                    'stock' => []];
            
            $currentSituation[$id_product]['stock'][$id_product_attribute] = $qty;
            $currentSituation[$id_product]['attr_prices'][$id_product_attribute] = $attrProductPrice;
        }

        // On va maintenant comparer les stocks et les mettre à jour lorsqu'ils sont différents
        foreach ($refs as $id_product_attribute => $stock)
        {
            $nbDone++;

            $id_product = $stock->id_product;
            if (!isset($currentStocks[$id_product_attribute]))
                $currentStocks[$id_product_attribute] = []; // Aucune info de stock pour le moment
  
            foreach ($stock->qties as $store_code => $qty)
            {
                if (!isset($currentStocks[$id_product_attribute][$store_code]))
                    $currentStocks[$id_product_attribute][$store_code] = 0; // pas de stock enregistré !

                if ($currentStocks[$id_product_attribute][$store_code] != $qty)
                {
                    // Stock différent => on le sauve
                    $currentStocks[$id_product_attribute][$store_code] = $qty;
                    $sql = 'INSERT INTO `'.$this->syncer->module->TblPrefix.'stocks` (`id_product`, `id_product_attribute`, `store_code`, `qty`) VALUES ('.$id_product.', ' . $id_product_attribute . ', \''.$store_code.'\', '.$qty.') ON DUPLICATE KEY UPDATE `id_product` = ' . $id_product . ', `qty` = '.$qty;
                    $res = $db->execute($sql);
                    if ($res === FALSE)
                        throw new Exception('Erreur lors de la mise à jour du stock : '.$db->getMsgError());                    
                }
            }

            // Ok, on va recalculer le stock global de la déclinaison
            $afterQty = 0;
            foreach ($currentStocks[$id_product_attribute] as $qty)
                $afterQty += $qty;

            // Application du stock de sécurité !
            $afterQty -= $securityStock;
            if ($afterQty < 0) $afterQty = 0;

            // Stock actuel
            $beforeStock = isset($currentSituation[$id_product]['stock'][$id_product_attribute]) ? 
                                $currentSituation[$id_product]['stock'][$id_product_attribute] : 0;

            // Mise à jour du stock
            if ($beforeStock != $afterQty)
            {
                StockAvailable::setQuantity($id_product, $id_product_attribute, $afterQty);
                $this->syncer->audit("Stock article #".$id_product_attribute." : ".$beforeStock." → ".$afterQty."");

                // Nouvelle situation
                $currentSituation[$id_product]['stock'][$id_product_attribute] = $afterQty;

                // Contrôle de la visibilité du produit
                if ($noStockVisibility)
                {
                    // Totalisation
                    $totalQty = 0;
                    foreach ($currentSituation[$id_product]['stock'] as $qty)
                        $totalQty += $qty;

                    // On va maintenant vérifier si le produit doit être caché ou non
                    $mustHaveVisibility = $totalQty > 0 ? 'both' : $noStockVisibility;

                    if (!isset($currentSituation[$id_product]['visibility']) || $currentSituation[$id_product]['visibility'] != $mustHaveVisibility)
                    {
                        // Visibilité différente => on la sauve
                        $p = new \Product($id_product);
                        if ($p->id == $id_product)
                        {
                            $p->visibility = $mustHaveVisibility;
                            $p->update();

                            // sauvegarde de l'info dans l'audit
                            $this->syncer->audit("Visibilité du produit ".$id_product." → '".$mustHaveVisibility."'");
                            $currentSituation[$id_product]['visibility'] = $p->visibility;
                        }
                    }
                }
            }

            // Contrôle des prix
            if ($stock->prixTTC !== null)
            {
                $currentProductPriceWT = $currentSituation[$id_product]['price'];
                if (isset($currentProductPriceWT))
                {
                    $newPriceWT = round($stock->prixTTC / (1 + $currentSituation[$id_product]['tax_rate'] / 100.0), 6);

                    $currentAttributePriceWT = $currentSituation[$id_product]['attr_prices'][$id_product_attribute] ?? 0;
                    $totalPriceWT = $currentProductPriceWT + $currentAttributePriceWT;

                    if ($newPriceWT != $totalPriceWT)
                    {
                        // Prix différent, on calcule la différence
                        $diff = $newPriceWT - $totalPriceWT;

                        // Nous devons changer le prix du product attribute
                        $this->syncer->audit("Changement de prix pour le produit #".$id_product." / #".$id_product_attribute." : ".$totalPriceWT." (ht) → ".$newPriceWT." (ht)");
                        $c = new \Combination($id_product_attribute);
                        if ($c->id == $id_product_attribute)
                        {
                            $c->price = $currentAttributePriceWT + $diff;
                            $c->update();
                        }
                    }
                };
            }
        }

        return $nbDone;
    }
}